import functools
import logging
from collections import Counter
from typing import TYPE_CHECKING, Callable, List, Optional, Tuple, Type, Union

from pyminion.core import (CardType, Card, Deck)
from pyminion.exceptions import (InsufficientMoney, InvalidBinaryInput,
                                 InvalidMultiCardInput, InvalidMultiOptionInput,
                                 InvalidSingleCardInput, InvalidDeckPositionInput)
from pyminion.player import Player

if TYPE_CHECKING:
    from pyminion.game import Game


logger = logging.getLogger()


def validate_input(
    func: Optional[Callable] = None,
    exceptions: Union[Tuple[Type[Exception], ...], Type[Exception]] = (),
):
    """
    Decorator to ensure that a user enters valid input when prompted.
    If input is invalid, the error is printed to the terminal and the user is prompted again

    Accepts a tuple of exceptions to catch

    """
    assert callable(func) or func is None

    def decorator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            while True:
                try:
                    return func(*args, **kwargs)
                except exceptions as e:
                    logger.error(e)

        return wrapper

    return decorator(func) if callable(func) else decorator


def binary_decision(prompt: str) -> bool:
    """
    Get user response to a binary question.
    Raise exception is input is anything other than 'yes' or 'no'

    """

    decision = input(prompt)
    if decision in ("yes", "y", "Yes", "Y"):
        return True
    elif decision in ("no", "n", "No", "N"):
        return False
    else:
        raise InvalidBinaryInput("Invalid response, valid choices are 'yes' or 'no'")


def single_card_decision(
    prompt: str, valid_cards: List[Card], valid_mixin: str = "placeholder"
) -> Optional[Union[Card, str]]:
    """
    Get user input when given the option to select one card

    valid_cards are a list of cards that a user can choose from. For example,
    when prompting a user to discard a card from their hand, valid cards would
    be the user's hand.

    valid_mixin allows for options other than a card to be selected. For example,
    "all" mixin would be used to autoplay treasures when prompting treasures to play

    Raise exception if user provided selection is not in valid_cards and is not
    a valid_mixin.

    """
    card_input = input(prompt)
    if not card_input:
        return None

    if card_input == valid_mixin:
        return valid_mixin

    for card in valid_cards:
        if card_input.casefold() == card.name.casefold():
            return card

    raise InvalidSingleCardInput(
        f"Invalid input, {card_input} is not a valid selection"
    )


def multiple_card_decision(
    prompt: str,
    valid_cards: List[Card],
    allow_all: bool = False,
) -> List[Card]:
    """
    Get user input when given the option to select multiple cards

    valid_cards are a list of cards that a user can choose from. For example,
    when prompting a user to trash cards from their hand, valid cards would
    be the user's hand.

    allow_all is a flag that, when True, allows the user to input "all" to
    return all valid cards.

    Raise exception if user provided selection is not in valid_cards.

    """
    card_input = input(prompt)
    if not card_input:
        return []

    if allow_all and card_input == "all":
        return valid_cards

    card_strings = [x.strip() for x in card_input.split(",")]
    selected_cards = []
    for card_string in card_strings:
        for card in valid_cards:
            if card_string.casefold() == card.name.casefold():
                selected_cards.append(card)
                break
        else:
            raise InvalidMultiCardInput(
                f"Invalid input, {card_string} is not a valid card"
            )

    selected_count = Counter(selected_cards)
    valid_count = Counter(valid_cards)

    for element in selected_count:
        if selected_count[element] > valid_count[element]:
            raise InvalidMultiCardInput(
                f"Invalid input, attempted to select too many copies of {element}"
            )
    return selected_cards


def multiple_option_decision(
    options: List[str],
    num_choices: int = 1,
    unique: bool = True,
) -> List[int]:
    """
    Get user input when given multiple options

    options are a list of strings describing the options that a user can choose
    from.

    Raise exception if user provided selection is not a valid option

    """
    prompt = f"Choose {num_choices} of the following"
    if unique and num_choices > 1:
        prompt += " (the choices must be different)"
    prompt += ":"
    print(prompt)
    for i, option in enumerate(options):
        print(f"{i + 1}: {option}")
    choice_input = input("Choice: ")
    choice_strings = [x.strip() for x in choice_input.split(",")]

    if len(choice_strings) != num_choices:
        raise InvalidMultiOptionInput("Invalid input, chose incorrect number of options")

    choices: List[int] = []
    for choice in choice_strings:
        try:
            choice_num = int(choice)
        except ValueError:
            raise InvalidMultiOptionInput(f"'{choice}' is not a valid number")

        if choice_num <= 0 or choice_num > len(options):
            raise InvalidMultiOptionInput(f"'{choice_num}' is not a valid option")

        choice_index = choice_num - 1
        if unique and choice_index in choices:
            raise InvalidMultiOptionInput("Choices are not unique")

        choices.append(choice_index)

    return choices


def deck_position_decision(prompt: str, num_deck_cards: int) -> int:
    pos_str = input(prompt).casefold()
    if pos_str == "top":
        pos = num_deck_cards
    elif pos_str == "bottom":
        pos = 0
    else:
        try:
            num = int(pos_str)
        except ValueError:
            raise InvalidDeckPositionInput(f"Invalid input, '{pos_str}' is not a valid position")

        max_num = num_deck_cards + 1
        if num < 1:
            raise InvalidDeckPositionInput(f"Invalid input, '{pos_str}' is less than the minimum value of 1")
        elif num > max_num:
            raise InvalidDeckPositionInput(f"Invalid input, '{pos_str}' is greater than the maximum value of {max_num}")

        pos = num - 1

    return pos


class HumanDecider:
    """
    Prompts human for decisions through the terminal.

    """

    @validate_input(exceptions=InvalidSingleCardInput)
    def action_phase_decision(
        self,
        valid_actions: List["Card"],
        player: "Player",
        game: "Game",
    ) -> Optional["Card"]:
        card = single_card_decision(
            prompt="Choose an action card to play: ",
            valid_cards=valid_actions,
        )
        if isinstance(card, str):
            raise InvalidSingleCardInput("You must choose a valid card")

        return card

    @validate_input(exceptions=InvalidMultiCardInput)
    def treasure_phase_decision(
        self,
        valid_treasures: List["Card"],
        player: "Player",
        game: "Game",
    ) -> List["Card"]:
        cards = multiple_card_decision(
            prompt="Choose treasures to play: ",
            valid_cards=valid_treasures,
            allow_all=True,
        )
        return cards

    @validate_input(exceptions=InvalidSingleCardInput)
    def buy_phase_decision(
        self,
        valid_cards: List["Card"],
        player: "Player",
        game: "Game",
    ) -> Optional["Card"]:
        card = single_card_decision(
            prompt="Choose a card to buy: ",
            valid_cards=valid_cards,
        )
        if isinstance(card, str):
            raise InvalidSingleCardInput("You must choose a valid card")

        return card

    @validate_input(exceptions=InvalidBinaryInput)
    def binary_decision(
        self,
        prompt: str,
        card: Card,
        player: "Player",
        game: "Game",
        relevant_cards: Optional[List[Card]] = None,
    ) -> bool:
        """
        Wrap binary_decision with @validate_input decorator to
        repeat prompt if input is invalid.

        """
        return binary_decision(prompt=prompt)

    @validate_input(exceptions=InvalidMultiOptionInput)
    def multiple_option_decision(
        self,
        card: "Card",
        options: List[str],
        player: "Player",
        game: "Game",
        num_choices: int = 1,
        unique: bool = True,
    ) -> List[int]:
        """
        Wrap multiple_option_decision with @validate_input decorator to
        repeat prompt if input is invalid.

        """
        return multiple_option_decision(options, num_choices, unique)

    @validate_input(exceptions=InvalidMultiCardInput)
    def discard_decision(
        self,
        prompt: str,
        card: "Card",
        valid_cards: List["Card"],
        player: "Player",
        game: "Game",
        min_num_discard: int = 0,
        max_num_discard: int = -1,
    ) -> List["Card"]:

        result = multiple_card_decision(prompt, valid_cards)
        len_result = len(result)

        if len_result < min_num_discard:
            raise InvalidMultiCardInput(
                f"Invalid response, you must discard at least {min_num_discard} card(s)"
            )
        elif max_num_discard >= 0 and len_result > max_num_discard:
            raise InvalidMultiCardInput(
                f"Invalid response, you cannot discard more than {max_num_discard} card(s)"
            )

        return result

    @validate_input(exceptions=InvalidMultiCardInput)
    def trash_decision(
        self,
        prompt: str,
        card: "Card",
        valid_cards: List["Card"],
        player: "Player",
        game: "Game",
        min_num_trash: int = 0,
        max_num_trash: int = -1,
    ) -> List["Card"]:

        result = multiple_card_decision(prompt, valid_cards)
        len_result = len(result)

        if len_result < min_num_trash:
            raise InvalidMultiCardInput(
                f"Invalid response, you must trash at least {min_num_trash} card(s)"
            )
        elif max_num_trash >= 0 and len_result > max_num_trash:
            raise InvalidMultiCardInput(
                f"Invalid response, you cannot trash more than {max_num_trash} card(s)"
            )

        return result

    @validate_input(exceptions=InvalidMultiCardInput)
    def gain_decision(
        self,
        prompt: str,
        card: "Card",
        valid_cards: List["Card"],
        player: "Player",
        game: "Game",
        min_num_gain: int = 0,
        max_num_gain: int = -1,
    ) -> List["Card"]:
        result = multiple_card_decision(prompt, valid_cards)
        len_result = len(result)

        if len_result < min_num_gain:
            raise InvalidMultiCardInput(
                f"Invalid response, you must gain at least {min_num_gain} card(s)"
            )
        elif max_num_gain >= 0 and len_result > max_num_gain:
            raise InvalidMultiCardInput(
                f"Invalid response, you cannot gain more than {max_num_gain} card(s)"
            )

        return result

    @validate_input(exceptions=InvalidMultiCardInput)
    def topdeck_decision(
        self,
        prompt: str,
        card: "Card",
        valid_cards: List["Card"],
        player: "Player",
        game: "Game",
        min_num_topdeck: int = 0,
        max_num_topdeck: int = -1,
    ) -> List["Card"]:
        result = multiple_card_decision(prompt, valid_cards)
        len_result = len(result)

        if len_result < min_num_topdeck:
            raise InvalidMultiCardInput(
                f"Invalid response, you must topdeck at least {min_num_topdeck} card(s)"
            )
        elif max_num_topdeck >= 0 and len_result > max_num_topdeck:
            raise InvalidMultiCardInput(
                f"Invalid response, you cannot topdeck more than {max_num_topdeck} card(s)"
            )

        return result

    @validate_input(exceptions=InvalidDeckPositionInput)
    def deck_position_decision(
        self,
        prompt: str,
        card: "Card",
        player: "Player",
        game: "Game",
        num_deck_cards: int,
    ) -> int:
        """
        Wrap deck_position_decision with @validate_input decorator to
        repeat prompt if input is invalid.

        """
        return deck_position_decision(prompt, num_deck_cards)

    @validate_input(exceptions=InvalidMultiCardInput)
    def reveal_decision(
        self,
        prompt: str,
        card: "Card",
        valid_cards: List["Card"],
        player: "Player",
        game: "Game",
        min_num_reveal: int = 0,
        max_num_reveal: int = -1,
    ) -> List["Card"]:
        result = multiple_card_decision(prompt, valid_cards)
        len_result = len(result)

        if len_result < min_num_reveal:
            raise InvalidMultiCardInput(
                f"Invalid response, you must reveal at least {min_num_reveal} card(s)"
            )
        elif max_num_reveal >= 0 and len_result > max_num_reveal:
            raise InvalidMultiCardInput(
                f"Invalid response, you cannot reveal more than {max_num_reveal} card(s)"
            )

        return result

    @validate_input(exceptions=InvalidMultiCardInput)
    def pass_decision(
        self,
        prompt: str,
        card: "Card",
        valid_cards: List["Card"],
        player: "Player",
        game: "Game",
        min_num_pass: int = 0,
        max_num_pass: int = -1,
    ) -> List["Card"]:
        result = multiple_card_decision(prompt, valid_cards)
        len_result = len(result)

        if len_result < min_num_pass:
            raise InvalidMultiCardInput(
                f"Invalid response, you must pass at least {min_num_pass} card(s)"
            )
        elif max_num_pass >= 0 and len_result > max_num_pass:
            raise InvalidMultiCardInput(
                f"Invalid response, you cannot pass more than {max_num_pass} card(s)"
            )

        return result

    @validate_input(exceptions=InvalidMultiCardInput)
    def name_card_decision(
        self,
        prompt: str,
        card: "Card",
        valid_cards: List["Card"],
        player: "Player",
        game: "Game",
        min_num_name: int = 0,
        max_num_name: int = -1,
    ) -> List["Card"]:
        result = multiple_card_decision(prompt, valid_cards)
        len_result = len(result)

        if len_result < min_num_name:
            raise InvalidMultiCardInput(
                f"Invalid response, you must name at least {min_num_name} card(s)"
            )
        elif max_num_name >= 0 and len_result > max_num_name:
            raise InvalidMultiCardInput(
                f"Invalid response, you cannot name more than {max_num_name} card(s)"
            )

        return result

    @validate_input(exceptions=InvalidSingleCardInput)
    def multi_play_decision(
        self,
        prompt: str,
        card: "Card",
        valid_cards: List["Card"],
        player: "Player",
        game: "Game",
        required: bool = True,
    ) -> Optional["Card"]:
        result = single_card_decision(prompt, valid_cards)

        if isinstance(result, str):
            raise InvalidSingleCardInput(
                f"Invalid response, you must name a valid card"
            )
        elif required and result is None:
            raise InvalidSingleCardInput(
                f"Invalid response, you must name a valid card"
            )

        return result


class Human(Player):
    """
    Human player can make choices as to which cards
    to play and buy in real time through the terminal

    """

    def __init__(
        self,
        deck: Optional[Deck] = None,
        player_id: str = "human",
    ):
        super().__init__(decider=HumanDecider(), deck=deck, player_id=player_id)



from pyminion.bots.bot import Bot, BotDecider
from pyminion.expansions.base import gold, province, silver, smithy
from pyminion.player import Player
from pyminion.game import Game


class BigMoneySmithyDecider(BotDecider):
    """
    Big money + smithy

    """

    def action_priority(self, player: Player, game: Game):
        yield smithy

    def buy_priority(self, player: Player, game: Game):
        money = player.state.money
        if money >= 8:
            yield province
        if money >= 6:
            yield gold
        if money == 4:
            yield smithy
        if money >= 3:
            yield silver


class BigMoneySmithy(Bot):
    def __init__(
        self,
        player_id: str = "big_money_smithy",
    ):
        super().__init__(decider=BigMoneySmithyDecider(), player_id=player_id)



import copy
import logging
from typing import Dict, List

from pyminion.game import Game
from pyminion.player import Player
from pyminion.result import GameResult, PlayerSimulatorResult, SimulatorResult

logger = logging.getLogger()


def get_percent(occurrence: int, total: int) -> float:
    """
    helper function to compute percent

    """
    return round(((occurrence / total) * 100), 3)


class Simulator:
    """
    Simulate multiple games of dominion and compute statistics

    Attributes:
        game: pyminion game instance.
        iterations: number of times the game will be simulated.

    """

    def __init__(self, game: Game, iterations: int = 100):
        self.game = game
        self.iterations = iterations
        self.results: List[GameResult] = []

    def run(self) -> SimulatorResult:
        logger.info(f"Simulating {self.iterations} games...")
        for _ in range(self.iterations):
            game = copy.copy((self.game))
            result = game.play()
            self.results.append(result)

        return self.get_sim_result()

    def get_sim_result(self) -> SimulatorResult:

        # make temp hashmap to store player sim results
        player_results: Dict[Player, PlayerSimulatorResult] = {}

        # initialize each player result with default values
        for player in self.game.players:
            player_results[player] = PlayerSimulatorResult(
                player=player, wins=0, losses=0, ties=0
            )

        # iterate through each simulated game to determine win record
        for result in self.results:

            # single player wins
            if len(result.winners) == 1:
                player_results[result.winners[0]].wins += 1

            # multiple players tie
            else:
                for player in result.winners:
                    player_results[player].ties += 1

            # rest of players are losers
            for player in self.game.players:
                if player not in result.winners:
                    player_results[player].losses += 1

        player_results_final: List[PlayerSimulatorResult] = list(
            player_results.values()
        )

        sim_result = SimulatorResult(
            iterations=self.iterations,
            game_results=self.results,
            player_results=player_results_final,
        )
        return sim_result



__version__ = "0.3.0"
__author__ = "Evan Slack"

import logging

# initalize logger with no handler.
# handlers are added in the `Game` init
logger = logging.getLogger()
logger.setLevel((logging.INFO))
logger.addHandler(logging.NullHandler())



import logging
import random
from collections import Counter
from typing import TYPE_CHECKING, Any, Iterable, Iterator, List, Optional, Tuple

if TYPE_CHECKING:
    from pyminion.game import Game
    from pyminion.player import Player

from enum import Enum
from pyminion.exceptions import EmptyPile, InsufficientActions, PileNotFound

logger = logging.getLogger()



class CardType(Enum):
    """
    Enum class for all card types that are currently used in the implemented expansions

    """
    Treasure = 1
    Victory = 2
    Curse = 3
    Action = 4
    Attack = 5
    Reaction = 6

class Card:

    """
    Base class representing a dominion card

    """

    def __init__(self, name: str, cost: int, type: Tuple[CardType]):
        self.name = name
        self._cost = cost
        self.type = type

    def __repr__(self):
        return f"{self.name}"

    def get_cost(self, player: "Player", game: "Game") -> int:
        cost = max(0, self._cost - game.card_cost_reduction)
        return cost


class Victory(Card):
    def __init__(self, name: str, cost: int, type: Tuple[CardType]):
        super().__init__(name, cost, type)

    def score(self, player: "Player") -> int:
        """
        Specific score method unique to each victory card

        """
        raise NotImplementedError(f"Score method must be implemented for {self.name}")


class Treasure(Card):
    def __init__(self, name: str, cost: int, type: Tuple[CardType], money: int):
        super().__init__(name, cost, type)
        self.money = money

    def play(self, player: "Player"):
        """
        Specific play method unique to each treasure card

        """
        raise NotImplementedError(f"Play method must be implemented for {self.name}")


class Action(Card):
    def __init__(
        self,
        name: str,
        cost: int,
        type: Tuple[CardType],
        actions: int = 0,
        draw: int = 0,
        money: int = 0,
    ):
        super().__init__(name, cost, type)
        self.actions = actions
        self.draw = draw
        self.money = money

    def play(self, player: "Player", game: "Game", generic_play: bool = True) -> None:
        """
        Specific play method unique to each action card

        """
        raise NotImplementedError(f"play method must be implemented for {self.name}")

    def generic_play(self, player: "Player") -> None:
        """
        Generic play method that gets executes for all action cards

        """
        if player.state.actions < 1:
            raise InsufficientActions(
                f"{player.player_id}: Not enough actions to play {self.name}"
            )

        player.playmat.add(self)
        player.hand.remove(self)
        player.state.actions -= 1

    def multi_play(self, player: "Player", game: "Game", state: Any, generic_play: bool = True) -> Any:
        """
        Called by "Throne Room variants" to play a card multiple times.
        By default this just calls self.play() but can be overridden by derived
        cards to implement different behavior.

        """
        self.play(player, game, generic_play)
        return None


class DeckCounter(Counter):
    def __str__(self):
        return ", ".join(f"{value} {key}" for key, value in (self).items())


class AbstractDeck:

    """
    Base class representing a generic list of dominion cards

    """

    def __init__(self, cards: Optional[List[Card]] = None):
        if cards:
            self.cards = cards
        else:
            self.cards = []

    def __repr__(self):
        return str(DeckCounter(self.cards))

    def __len__(self):
        return len(self.cards)

    def add(self, card: Card) -> None:
        self.cards.append(card)

    def remove(self, card: Card) -> Card:
        self.cards.remove(card)
        return card

    def move_to(self, destination: "AbstractDeck"):
        destination.cards += self.cards
        self.cards = []


class Deck(AbstractDeck):
    def __init__(self, cards: Optional[List[Card]] = None):
        super().__init__(cards)

    def draw(self) -> Card:
        drawn_card = self.cards.pop()
        return drawn_card

    def shuffle(self):
        random.shuffle(self.cards)


class DiscardPile(AbstractDeck):
    def __init__(self, cards: Optional[List[Card]] = None):
        super().__init__(cards)


class Hand(AbstractDeck):
    def __init__(self, cards: Optional[List[Card]] = None):
        super().__init__(cards)


class Pile(AbstractDeck):
    def __init__(self, cards: Optional[List[Card]] = None):
        super().__init__(cards)
        if cards and len(set(cards)) == 1:
            self.name = cards[0].name
        elif cards:
            self.name = "Mixed"
        else:
            self.name = None

    def remove(self, card: Card) -> Card:
        if len(self.cards) < 1:
            raise EmptyPile(f"{self.name} pile is empty, cannot gain card")
        self.cards.remove(card)
        return card


class Playmat(AbstractDeck):
    def __init__(self, cards: Optional[List[Card]] = None):
        super().__init__(cards)


class Trash(AbstractDeck):
    def __init__(self, cards: Optional[List[Card]] = None):
        super().__init__(cards)


class Supply:
    """
    Collection of card piles that make up the game's supply.

    """

    def __init__(self, piles: Optional[List[Pile]] = None):
        if piles:
            self.piles = piles
        else:
            self.piles = []

    def __repr__(self):
        return str(self.avaliable_cards())

    def __len__(self):
        return len(self.piles)

    def get_pile(self, pile_name: str) -> Pile:
        """
        Get a pile by name.

        """
        for pile in self.piles:
            if pile.name == pile_name:
                return pile
        raise PileNotFound(f"{pile_name} pile is not valid")

    def gain_card(self, card: Card) -> Card:
        """
        Gain a card from the supply.

        """
        pile = self.get_pile(card.name)
        try:
            return pile.remove(card)

        except EmptyPile as e:
            raise e

    def return_card(self, card: Card):
        """
        Return a card to the supply.

        """
        pile = self.get_pile(card.name)
        pile.add(card)

    def trash_card(self, card: Card, trash: "Trash") -> None:
        """
        Trash a card from the supply.

        """
        pile = self.get_pile(card.name)
        pile.remove(card)
        trash.add(card)

    def avaliable_cards(self) -> List[Card]:
        """
        Returns a list containing a single card from each non-empty pile in the supply.

        """
        cards = [pile.cards[0] for pile in self.piles if pile]
        return cards

    def num_empty_piles(self) -> int:
        """
        Returns the number of empty piles in the supply.

        """
        empty_piles: int = 0
        for pile in self.piles:
            if len(pile) == 0:
                empty_piles += 1
        return empty_piles

    def pile_length(self, pile_name: str) -> int:
        """
        Get the number of cards in a specified pile in the supply.

        """
        pile = self.get_pile(pile_name)
        return len(pile)


def get_action_cards(cards: Iterable[Card]) -> Iterator[Action]:
    for card in cards:
        if CardType.Action in card.type:
            assert isinstance(card, Action)
            yield card


def get_treasure_cards(cards: Iterable[Card]) -> Iterator[Treasure]:
    for card in cards:
        if CardType.Treasure in card.type:
            assert isinstance(card, Treasure)
            yield card


def get_victory_cards(cards: Iterable[Card]) -> Iterator[Victory]:
    for card in cards:
        if CardType.Victory in card.type:
            assert isinstance(card, Victory)
            yield card



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="pyminion",
    version="0.4.0",
    author="Evan Slack",
    author_email="evan.slack@outlook.com",
    description="Dominion but make it python",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/evanofslack/pyminion",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    packages=setuptools.find_packages(
        exclude=(
            "tests",
            "tests.*",
            "examples",
            "examples.*",
        )
    ),
    python_requires=">=3.8",
)
